#! /usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import logging
import os
import sys
from collections import defaultdict
import configargparse

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
import camsa.utils.ragout.io as ragout_io
from camsa.utils.ragout.shared import filter_indels, filter_duplications
from camsa.utils.ragout.shared import filter_blocks_by_good_genomes, filter_blocks_by_bad_genomes, get_all_genomes_from_blocks

if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Computing coverage report for Ragout blocks.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "logging.ini")])
    parser.add_argument("-c", "--config", is_config_file=True, help="Config file overwriting some of the default settings as well as any flag starting with \"--\".")

    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("ragout_coords", type=str, help="A path to ragout coords file")
    parser.add_argument("--filter-indels", action="store_true", dest="filter_indels", default=False)
    parser.add_argument("--no-fragment-stats", action="store_false", dest="fragment_stats", default=True)
    parser.add_argument("--no-genome-stats", action="store_false", dest="genome_stats", default=True)
    parser.add_argument("--filter-duplications", action="store_true", dest="filter_duplications", default=False)
    parser.add_argument("--good-genomes", type=str, default="", help="A coma separated list of genome names, to be processed and conversed.\nDEFAULT: \"\" (i.e., all genomes are good)")
    parser.add_argument("--bad-genomes", type=str, default="", help="A coma separated list of genome names, to be excluded from processing and conversion.\nDEFAULT: \"\" (i.e., no genomes are bad)")
    parser.add_argument("-o", "--output", type=configargparse.FileType("wt"), default=sys.stdout)
    parser.add_argument("--c-logging-level", dest="c_logging_level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    args = parser.parse_args()

    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.ragout_coords2fasta")
    ch = logging.StreamHandler()
    ch.setLevel(args.c_logging_level)
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    sequences_by_ids, blocks_by_ids = ragout_io.read_from_file(path=args.ragout_coords, silent_fail=False, delimiter="\t")
    all_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.good_genomes != "":
        args.good_genomes = set(args.good_genomes.split(","))
        filter_blocks_by_good_genomes(blocks_by_ids=blocks_by_ids, good_genomes=args.good_genomes)
    if args.bad_genomes != "":
        args.bad_genomes = set(args.bad_genomes.split(","))
        filter_blocks_by_bad_genomes(blocks_by_ids=blocks_by_ids, bad_genomes=args.bad_genomes)
    if args.filter_indels:
        filter_indels(blocks_by_ids=blocks_by_ids, all_genomes_as_set=(all_genomes if len(args.good_genomes) == 0 else args.good_genomes) - args.bad_genomes)
    if args.filter_duplications:
        filter_duplications(blocks_by_ids=blocks_by_ids)
    all_filtered_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)

    genomes = defaultdict(lambda: defaultdict(list))
    for block_list in blocks_by_ids.values():
        for block in block_list:
            genomes[block.parent_seq.genome_name][block.parent_seq.ragout_id].append(block)

    fragment_cov = {}
    if args.fragment_stats:
        for genome_name in genomes.keys():
            for seq_id in genomes[genome_name]:
                seq = sequences_by_ids[seq_id]
                cumulative_blocks_length = sum(block.length for block in genomes[genome_name][seq_id])
                fragment_cov[seq_id] = cumulative_blocks_length * 100.0 / seq.length

    genome_cov = {}
    if args.genome_stats:
        for genome_name in genomes.keys():
            total_genome_length = 0
            total_blocks_length = 0
            for seq_id in genomes[genome_name]:
                seq = sequences_by_ids[seq_id]
                total_genome_length += seq.length
                total_blocks_length += sum(block.length for block in genomes[genome_name][seq_id])
            genome_cov[genome_name] = total_blocks_length * 100.0 / total_genome_length

    if args.genome_stats:
        print("-" * 80, file=args.output)
        for genome_name in sorted(genomes.keys()):
            print("For genome \"{genome_name}\" {cov:.2f}% of its length is covered by filtered blocks".format(genome_name=genome_name, cov=genome_cov[genome_name]), file=args.output)
    if args.fragment_stats:
        for genome_name in sorted(genomes.keys()):
            print("-"*80, file=args.output)
            print("Detailed coverage stats for fragments in genome \"{genome_name}\"".format(genome_name=genome_name), file=args.output)
            for seq_id in sorted(genomes[genome_name].keys()):
                print("For fragment \"{fragment_name}\" {cov:.2f}% of its length is covered by filtered blocks".format(fragment_name=sequences_by_ids[seq_id].seq_name, cov=fragment_cov[seq_id]))

    logger.info("All done!")
    end_time = datetime.datetime.now()
    logger.info("Elapsed time: {el_time}".format(el_time=str(end_time - start_time)))


# -*- coding: utf-8 -*-
from Bio.Seq import Seq

from camsa.utils.fasta.algo import bounded_alignment


class GapFilling(object):
    def __init__(self, seq):
        self.seq = seq
        self.score = None

    def suitable_to_fill_the_gap(self, gap_size, gap_size_error):
        return False

    def compute_score(self, gap_size, gap_size_error):
        non_N_content = len([letter for letter in self.seq if letter != "N"])
        self.score = non_N_content


class IntraGapFilling(GapFilling):
    def suitable_to_fill_the_gap(self, gap_size, gap_size_error):
        if gap_size == "?":
            return True
        min_gap_size = gap_size - gap_size_error
        max_gap_size = gap_size + gap_size_error
        return min_gap_size <= len(self.seq) <= max_gap_size


class FlankingGapFilling(GapFilling):
    def __init__(self, seq1, seq2, seq):
        super(FlankingGapFilling, self).__init__(seq)
        self.start_seq = seq1
        self.end_seq = seq2
        self.alignment_result = None

    def suitable_to_fill_the_gap(self, gap_size, gap_size_error, fill_remainder=True):
        if gap_size == "?":
            return False
        min_gap_size = gap_size - gap_size_error
        max_gap_size = gap_size + gap_size_error
        cumulative_length = len(self.start_seq) + len(self.end_seq)
        if cumulative_length < min_gap_size:
            return False
        if cumulative_length > gap_size + 2 * gap_size_error:
            return False
        if cumulative_length > max_gap_size:
            min_overlap = cumulative_length - gap_size - gap_size_error
            max_overlap = cumulative_length - gap_size + gap_size_error
            max_overlap = min(max_overlap, len(self.start_seq), len(self.end_seq))
            l1s = self.start_seq[-max_overlap:]
            l2p = self.end_seq[:max_overlap]
            alignment_result = bounded_alignment(seq1=l1s, seq2=l2p, max_edit_distance=max_overlap - min_overlap)
            if alignment_result.consensus is None:
                return False
            else:
                self.alignment_result = alignment_result
                return True
        if cumulative_length < max_gap_size:
            # there can be an overlap
            return True

    def prepare_seq(self, gap_size, gap_size_error):
        # only called if the gap can be suitable for filling w.r.t. gap size
        self.seq = self.start_seq
        cumulative_length = len(self.start_seq) + len(self.end_seq)
        if cumulative_length < gap_size + gap_size_error:
            self.seq = self.start_seq + self.end_seq
        else:
            max_overlap = cumulative_length - gap_size + gap_size_error
            max_overlap = min(max_overlap, len(self.start_seq), len(self.end_seq))
            self.seq = self.start_seq[:max_overlap] + self.alignment_result.consensus + self.end_seq[max_overlap:]



# -*- coding: utf-8 -*-
import itertools
import json
from collections import defaultdict

import networkx
import six
import sys


class AssemblyPoint(object):
    def __init__(self, seq1, seq2, seq1_or, seq2_or, sources, cw=None, parent_id=None, children_ids=None, self_id=None,
                 gap_size=None):
        self.seq1 = seq1
        self.seq2 = seq2
        self.seq1_or = seq1_or
        self.seq2_or = seq2_or
        self.sources = sorted(sources)
        self.cw = cw
        self.gap_size = gap_size
        self.self_id = self_id
        self.parent_id = parent_id
        self.children_ids = children_ids if children_ids is not None else []
        #### comparative metrics
        self.participates_in_merged = False
        self.in_conflicted = defaultdict(set)
        self.in_semi_conflicted = defaultdict(set)
        self.out_conflicted = defaultdict(set)
        self.out_semi_conflicted = defaultdict(set)
        self.seq1_par_or = None
        self.seq2_par_or = None
        #### ref metrics
        self.ref_metrics = RefMetrics()

    @property
    def orientation_as_word(self):
        if self.is_unoriented:
            return "U"
        if self.is_semi_oriented:
            return "SO"
        return "O"

    @property
    def is_unoriented(self):
        return [self.seq1_or, self.seq2_or].count("?") == 2

    @property
    def is_semi_oriented(self):
        return [self.seq1_or, self.seq2_or].count("?") == 1

    @property
    def is_oriented(self):
        return [self.seq1_or, self.seq2_or].count("?") == 0

    def is_in_semi_conflicted_for(self, source_name):
        return source_name in self.in_semi_conflicted

    def is_in_conflicted_for(self, source_name):
        return source_name in self.in_conflicted

    def is_out_semi_conflicted_for(self, source_name):
        return len(self.out_semi_conflicted) > 0

    def is_out_conflicted_for(self, source_name):
        return len(self.out_conflicted) > 0

    @property
    def is_out_conflicted(self):
        return len(self.out_conflicted) > 0

    @property
    def is_out_semi_conflicted(self):
        return len(self.out_semi_conflicted) > 0

    @property
    def is_non_conflicted(self):
        return len(self.in_conflicted) == 0 and len(self.out_conflicted) == 0 and len(self.in_semi_conflicted) == 0 and len(
            self.out_semi_conflicted) == 0

    def get_edges(self, weight=False, sort=True):
        result = []
        for ap in self.get_all_all_possible_realizations():
            head = ap.seq1 + ("h" if ap.seq1_or == "+" else "t")
            tail = ap.seq2 + ("t" if ap.seq2_or == "+" else "h")
            hv = head
            tv = tail
            if sort:
                hv, tv = (hv, tv) if hv < tv else (tv, hv)
            if weight:
                result.append((hv, tv, self.cw))
            else:
                result.append((hv, tv))
        return result

    def get_all_all_possible_realizations(self):
        result = []
        if self.seq1_or == "?":
            seq1_choices = ["+", "-"]
        else:
            seq1_choices = [self.seq1_or]
        if self.seq2_or == "?":
            seq2_choices = ["+", "-"]
        else:
            seq2_choices = [self.seq2_or]
        for ctg1_or, ctg2_or in itertools.product(seq1_choices, seq2_choices):
            result.append(AssemblyPoint(seq1=self.seq1, seq2=self.seq2, seq1_or=ctg1_or, seq2_or=ctg2_or,
                                        sources=self.sources))
        return result

    @property
    def is_ambiguous(self):
        return self.seq1_or == "?" or self.seq2_or == "?"


class APFieldOutExtractorConverter(object):
    def __init__(self, field_name, converter_name):
        self.field_name = field_name
        self.converter = AP_FIELD_CONVERTERS[converter_name]

    def extract_field_value_str(self, ap):
        return self.converter.convert(getattr(ap, self.field_name))


class APFieldConverter(object):
    @staticmethod
    def convert(field_value, **kwargs):
        raise NotImplemented("Converter must be field/area specific")


class APFieldConverterStr(APFieldConverter):
    @staticmethod
    def convert(field_value, **kwargs):
        return str(field_value)


class APFieldConverterQuestionIfNone(APFieldConverterStr):
    @staticmethod
    def convert(field_value, **kwargs):
        if field_value is None:
            field_value = "?"
        return super(APFieldConverterQuestionIfNone, APFieldConverterQuestionIfNone).convert(field_value=field_value, **kwargs)


class APFieldConverterBooleanToInt(APFieldConverterQuestionIfNone):
    @staticmethod
    def convert(field_value, **kwargs):
        if field_value in [True, False]:
            field_value = int(field_value)
        return super(APFieldConverterBooleanToInt, APFieldConverterBooleanToInt).convert(field_value=field_value, **kwargs)


IdFieldConverter = APFieldConverterQuestionIfNone


class IterableFieldConverter(APFieldConverterQuestionIfNone):
    @staticmethod
    def convert(field_value, **kwargs):
        result = []
        treat_as_set = kwargs.get("as_set", True)
        for value in field_value:
            result.append(super(IterableFieldConverter, IterableFieldConverter).convert(field_value=value, **kwargs))
        if treat_as_set:
            result = set(result)
        result = sorted(result)
        treat_empty_as_zero = kwargs.get("empty_as_zero", True)
        if treat_empty_as_zero and len(result) == 0:
            return "0"
        separator = kwargs.get("intra_separator", ",")
        return separator.join(result)


class ConflictFieldConverter(APFieldConverter):
    @staticmethod
    def convert(field_value):
        conflicted_ids = [ap_id for conflict_assembly in field_value.values() for ap_id in conflict_assembly]
        return IterableFieldConverter.convert(field_value=conflicted_ids)


class RefMetrics(object):
    def __init__(self):
        #### correct
        self.present = False
        self.present_ref_ids = []
        #### incorrect
        # the worst error
        self.interchromosomal = False
        # the best error
        self.tranlocation = False
        self.translocation_path = []
        # can be combined with translocation, or can be on its own
        self.reversal = False
        # a subtype of reversal, better than reversal. possible for semi/un-oriented assembly points only
        self.semi_reversal = False
        # a best available reading of the reference genome for this particular assembly point
        self.best_ref_reading_id = None


class MergedScaffoldAssemblyGraph(object):
    def __init__(self):
        self.graph = networkx.Graph()

    def add_edge(self, u, v, weight):
        if self.graph.has_edge(u=u, v=v):
            self.graph[u][v]['weight'] += weight
        else:
            self.graph.add_edge(u=u, v=v, weight=weight)

    def get_maximal_non_conflicting_assembly_graph(self):
        max_matching = networkx.max_weight_matching(G=self.graph)
        seen = set()
        edges = []
        for key, value in max_matching.items():
            if key not in seen and value not in seen:
                edges.append((key, value))
                seen.add(key)
                seen.add(value)
        result = networkx.Graph()
        result.add_edges_from(edges)
        return result

    def edges(self, weight=True):
        if weight:
            return [(u, v, w) if u < v else (v, u, w) for (u, v, w) in self.graph.edges(data='weight')]
        else:
            return [(u, v) if u < v else (v, u) for (u, v) in self.graph.edges()]

    def remove_edges_with_low_cw(self, cw_threshold=0.0):
        low_weight_edges = [(u, v) for (u, v, d) in self.graph.edges(data=True) if d['weight'] < cw_threshold]
        self.graph.remove_edges_from(low_weight_edges)


class ScaffoldAssemblyGraph(object):
    def __init__(self):
        self.graph = networkx.MultiGraph()

    def add_edge(self, u, v, **kwargs):
        self.graph.add_edge(u=u, v=v, attr_dict=kwargs)

    @classmethod
    def from_assembly_points(cls, assembly_points, reference=False, add_scaffold_edges=True):
        result = cls()
        scaffold_edges = get_scaffold_edges(assembly_points=assembly_points)
        if add_scaffold_edges:
            for u, v in scaffold_edges:
                result.add_edge(u=u, v=v, edge_type="scaffold")
        for ap in assembly_points:
            for (u, v, weight) in ap.get_edges(sort=True, weight=True):
                edge_type = "ref_assembly" if reference else "assembly"
                result.add_edge(u=u, v=v, weight=weight, edge_type=edge_type, ap_id=ap.self_id, ap_type=ap.orientation_as_word)
        return result


class Assembly(object):
    def __init__(self, name, aps):
        self.name = name
        self.aps = aps
        self.total_cnt = 0
        self.max_non_conflicting_aps = []

    @property
    def entries_names(self):
        return self.get_all_entries_names()

    def get_all_entries_names(self):
        return sorted(set(name for ap in self.aps for name in ap.sources))

    def sort_aps(self):
        self.aps = sorted(self.aps,
                          key=lambda ap: (ap.contig_1, ap.contig_1_orientation, ap.contig_2, ap.contig_2_orientation))

    @property
    def unoriented_ap_cnt(self):
        return sum(ap.is_unoriented for ap in self.aps)

    @property
    def semi_oriented_aps_cnt(self):
        return sum(ap.is_semi_oriented for ap in self.aps)

    @property
    def oriented_aps_cnt(self):
        return sum(ap.is_oriented for ap in self.aps)

    @property
    def in_conflicted_cnt(self):
        cnt = 0
        for ap in self.aps:
            if isinstance(self.name, tuple):
                to_check = self.name
            else:
                to_check = [self.name]
            for name in to_check:
                if ap.is_in_conflicted_for(name):
                    cnt += 1
                    break
        return cnt

    @property
    def in_semi_conflicted_cnt(self):
        cnt = 0
        for ap in self.aps:
            if isinstance(self.name, tuple):
                to_check = self.name
            else:
                to_check = [self.name]
            for name in to_check:
                if ap.is_in_semi_conflicted_for(name):
                    cnt += 1
                    break
        return cnt

    @property
    def out_conflicted_cnt(self):
        return sum(ap.is_out_conflicted for ap in self.aps)

    @property
    def out_semi_conflicted_cnt(self):
        return sum(ap.is_out_semi_conflicted for ap in self.aps)

    @property
    def non_conflicted_cnt(self):
        return len([ap for ap in self.aps if ap.is_non_conflicted])


def assign_ids_to_assembly_points(assembly_points, id_prefix="", id_generator=None):
    assembly_points_by_ids = {}
    if id_generator is None:
        id_generator = itertools.count()
    for assembly_point in assembly_points:
        assembly_point.self_id = id_prefix + str(six.next(id_generator))
        assembly_points_by_ids[assembly_point.self_id] = assembly_point
    return assembly_points_by_ids


def merge_assembly_points(assembly_points_by_source):
    unique_assembly_points = defaultdict(list)

    for assembly_points in assembly_points_by_source.values():
        for assembly_point in assembly_points:
            seq1, seq2 = assembly_point.seq1, assembly_point.seq2
            seq1_or, seq2_or = assembly_point.seq1_or, assembly_point.seq2_or
            if seq1 < seq2:
                entry = (seq1, seq1_or, seq2, seq2_or)
            else:
                entry = (seq2, inverse_orientation(seq2_or), seq1, inverse_orientation(seq1_or))
            unique_assembly_points[entry].append(assembly_point)
    result = []
    for (seq1, seq1_or, seq2, seq2_or), children in unique_assembly_points.items():
        sources = sorted(set(source for ap in children for source in ap.sources))
        weight = sum(ap.cw for ap in children)
        children_ids = [ap.self_id for ap in children]
        merged_assembly_point = AssemblyPoint(seq1=seq1, seq2=seq2, seq1_or=seq1_or, seq2_or=seq2_or,
                                              sources=sources, cw=weight, children_ids=children_ids)
        result.append(merged_assembly_point)
    return result


def assign_parents_to_children(children_assembly_points_by_ids, parent_assembly_points_by_ids):
    for p_assembly_point in parent_assembly_points_by_ids.values():
        for child_id in p_assembly_point.children_ids:
            children_assembly_points_by_ids[child_id].parent_id = p_assembly_point.self_id


def inverse_orientation(orientation):
    if orientation in ("+", "-"):
        return "+" if orientation == "-" else "-"
    return orientation


AP_FIELD_CONVERTERS = {
    "str_raw": APFieldConverterStr,
    "str": APFieldConverterQuestionIfNone,
    "bool": APFieldConverterBooleanToInt,
    "conflict": ConflictFieldConverter,
    "id": IdFieldConverter,
    "iter": IterableFieldConverter,
}


def get_scaffold_edges(assembly_points):
    unique_scaffolds = set()
    for ap in assembly_points:
        unique_scaffolds.add(ap.seq1)
        unique_scaffolds.add(ap.seq2)
    return [(name + "t", name + "h") for name in unique_scaffolds]


def to_json(value):
    if isinstance(value, Sequence):
        return json.dumps({"seq_id": value.name, "length": value.length})
    result = {}
    for key, values in value.items():
        if isinstance(values, str):
            result[key] = values
        else:
            result[key] = sorted(values)
    return json.dumps(result)


class Sequence(object):
    def __init__(self, name, length=None, parent_seq_id=None, start=None, end=None, strand="+", annotation=None, seq_group_id="Default"):
        self.name = name
        self._length = length
        self._parent_seq_id = parent_seq_id
        self._start = start
        self._end = end
        self._seq_group_id = seq_group_id
        self.strand = strand
        self._annotation = annotation

    @property
    def length(self):
        if self._length is not None:
            return self._length
        if self.start is not None and self.end is not None:
            return self.end - self.start
        return None

    @property
    def parent_seq_id(self):
        if self._parent_seq_id is None:
            return self.name
        return self._parent_seq_id

    @property
    def start(self):
        if self._start is None:
            return 0
        return int(self._start)

    @property
    def end(self):
        if self._end is None:
            return sys.maxsize
        return int(self._end)

    @property
    def seq_group_id(self):
        return str(self._seq_group_id)

    @property
    def annotation(self):
        return str(self._annotation)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import logging
import os
import sys
from collections import defaultdict

import configargparse
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
import camsa.utils.ragout.io as ragout_io
from camsa.utils.ragout.shared import filter_indels, filter_duplications
from camsa.utils.ragout.shared import filter_blocks_by_good_genomes, filter_blocks_by_bad_genomes, get_all_genomes_from_blocks

if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting Ragout formatted blocks into FASTA format.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "logging.ini")])
    parser.add_argument("-c", "--config", is_config_file=True, help="Config file overwriting some of the default settings as well as any flag starting with \"--\".")

    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("ragout_coords", type=str, help="A path to ragout coords file")
    parser.add_argument("--ann-genomes", type=str, default="")
    parser.add_argument("--ann-delimiter", type=str, default=";")
    parser.add_argument("--filter-indels", action="store_true", dest="filter_indels", default=False)
    parser.add_argument("--filter-duplications", action="store_true", dest="filter_duplications", default=False)
    parser.add_argument("--good-genomes", type=str, default="", help="A coma separated list of genome names, to be processed and conversed.\nDEFAULT: \"\" (i.e., all genomes are good)")
    parser.add_argument("--bad-genomes", type=str, default="", help="A coma separated list of genome names, to be excluded from processing and conversion.\nDEFAULT: \"\" (i.e., no genomes are bad)")
    parser.add_argument("fasta", nargs="+")
    parser.add_argument("-o", "--output", type=configargparse.FileType("wt"), default=sys.stdout)
    parser.add_argument("--o-genomes", dest="ref_genomes", type=str, default="a string of coma separated names of genomes, who will")
    parser.add_argument("--c-logging-level", dest="c_logging_level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    args = parser.parse_args()

    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.ragout_coords2fasta")
    ch = logging.StreamHandler()
    ch.setLevel(args.c_logging_level)
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    sequences_by_ids, blocks_by_ids = ragout_io.read_from_file(path=args.ragout_coords, silent_fail=False, delimiter="\t")
    all_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.good_genomes != "":
        args.good_genomes = set(args.good_genomes.split(","))
        filter_blocks_by_good_genomes(blocks_by_ids=blocks_by_ids, good_genomes=args.good_genomes)
    if args.bad_genomes != "":
        args.bad_genomes = set(args.bad_genomes.split(","))
        filter_blocks_by_bad_genomes(blocks_by_ids=blocks_by_ids, bad_genomes=args.bad_genomes)
    if args.filter_indels:
        filter_indels(blocks_by_ids=blocks_by_ids, all_genomes_as_set=(all_genomes if len(args.good_genomes) == 0 else args.good_genomes) - args.bad_genomes)
    if args.filter_duplications:
        filter_duplications(blocks_by_ids=blocks_by_ids)
    all_filtered_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.ref_genomes == "":
        logger.info("Output genomes were not specified")
        args.ref_genomes = sorted(all_filtered_genomes)
        logger.info("Setting output genomes to {ref_genome}".format(ref_genome=args.ref_genomes))
    else:
        args.ref_genomes = args.ref_genomes.split(",")
        for genome in args.ref_genomes:
            if genome not in all_filtered_genomes:
                logger.critical("Output genome {ref_genome} was not present in all filtered genomes {filtered_genomes}"
                                "".format(ref_genome=args.genome, filtered_genomes=",".join(all_filtered_genomes)))
                exit(1)
    if args.ann_genomes == "":
        args.ann_genomes = args.ref_genomes
    else:
        args.ann_genomes = args.ann_genomes.split(",")
        for genome in args.ann_genomes:
            if genome not in all_filtered_genomes:
                logger.critical("Annotation genomes {ann_genome} was not present in all filtered genomes {filtered_genomes}"
                                "".format(ann_genome=args.genome, filtered_genomes=",".join(all_filtered_genomes)))
                exit(1)
    blocks_to_convert = []
    for block_id in sorted(blocks_by_ids.keys()):
        blocks = blocks_by_ids[block_id]
        blocks_by_ref_genomes = [block for block in blocks if block.parent_seq.genome_name in args.ref_genomes]
        blocks = []
        for ref_genome in args.ref_genomes:
            for block in blocks_by_ref_genomes:
                if block.parent_seq.genome_name == ref_genome:
                    blocks.append(block)
        annotations = []
        for ann_genome in args.ann_genomes:
            for block in blocks_by_ref_genomes:
                if block.parent_seq.genome_name == ann_genome:
                    annotations.append(block.annotation_name)
        block = blocks[0]
        if len(annotations) != 0:
            block._annotation = args.ann_delimiter.join(annotations)
        blocks_to_convert.append(block)

    blocks_by_seq_ids = defaultdict(list)
    for block in blocks_to_convert:
        blocks_by_seq_ids[block.parent_seq.seq_name].append(block)

    processed = set()
    for f in args.fasta:
        logger.info("Processing fasta file: \"{file_name}\"".format(file_name=f))
        cnt = 0
        for record in SeqIO.parse(f, "fasta"):
            seq_id = record.id
            if seq_id not in blocks_by_seq_ids:
                continue
            current_blocks = blocks_by_seq_ids[seq_id]
            current_blocks = [block for block in current_blocks if block.name not in processed]
            for block in current_blocks:
                if block.strand == "+":
                    out_seq = record.seq[block.start: block.end]
                else:
                    out_seq = record.seq[block.start: block.end].reverse_complement()
                out_record = SeqRecord(seq=out_seq, id=str(block.name), description=block.annotation_name)
                SeqIO.write(sequences=out_record, handle=args.output, format="fasta")
                processed.add(block.name)

    logger.info("All done!")
    end_time = datetime.datetime.now()
    logger.info("Elapsed time: {el_time}".format(el_time=str(end_time - start_time)))



# -*- coding: utf-8 -*-
# noinspection PyCompatibility
import enum
import numbers
from collections import defaultdict

import blist
import networkx

from camsa.core.data_structures import MergedScaffoldAssemblyGraph, inverse_orientation
from camsa.core.data_structures import get_scaffold_edges


class MergingStrategies(enum.Enum):
    greedy_merging = "greedy"
    maximal_matching = "maximal-matching"


def get_un_oriented_assembly_points(assembly_points):
    return {ap for ap in assembly_points if ap.is_unoriented}


####################################################################
####################################################################
#                                                                  #
# Progressive merging strategy for combining scaffold assemblies   #
#                                                                  #
####################################################################

def merge_greedily(assembly_points_by_sources, acyclic=True, min_cw=0.0):
    def get_redundant_edges_from_assembly_points(e, points_by_edges, processed_points):
        result = []
        for assembly_point in points_by_edges[e]:
            if assembly_point in processed_points:
                continue
            processed_points.add(assembly_point)
            for ap_representation_edge in assembly_point.get_edges(sort=True):
                result.append(ap_representation_edge)
        return result

    assembly_points_by_sources = [ap for ap_list in assembly_points_by_sources.values() for ap in ap_list]
    already_processed_assembly_points = set()
    scaffold_edges = get_scaffold_edges(assembly_points=assembly_points_by_sources)
    cover_graph = networkx.Graph()
    cover_graph.add_edges_from(scaffold_edges)
    end_points = {}
    for (u, v) in cover_graph.edges_iter():
        end_points[u] = v
        end_points[v] = u

    merged_scaffold_assembly_edges_graph = MergedScaffoldAssemblyGraph()
    assembly_points_by_edges = defaultdict(list)
    for ap in assembly_points_by_sources:
        for (u, v, weight) in ap.get_edges(sort=True, weight=True):
            assembly_points_by_edges[(u, v)].append(ap)
            merged_scaffold_assembly_edges_graph.add_edge(u, v, weight=weight)
    merged_scaffold_assembly_edges_graph.remove_edges_with_low_cw(cw_threshold=min_cw)
    assembly_edges = merged_scaffold_assembly_edges_graph.edges(weight=True)
    sorted_assembly_edges = sorted(assembly_edges, key=lambda entry: entry[2])
    assembly_edges = blist.sortedlist(sorted_assembly_edges, key=lambda entry: entry[2])

    while len(assembly_edges) > 0:
        edge = assembly_edges.pop()
        redundant_edges = set()
        u, v, w = edge
        if (end_points[u] == v or end_points[v] == u) and acyclic:  # endpoints of the same path and graph is restrained to be acyclic
            redundant_edges.add((u, v))
        else:
            # growing the cover
            cover_graph.add_edge(u, v, weight=w)
            # updating endpoints of merged paths
            end_points[end_points[u]], end_points[end_points[v]] = end_points[v], end_points[u]
            # collecting redundant edges
            re1 = {tuple(sorted(e)) for e in merged_scaffold_assembly_edges_graph.graph.edges(nbunch=u)}
            re2 = {tuple(sorted(e)) for e in merged_scaffold_assembly_edges_graph.graph.edges(nbunch=v)}
            re3 = {tuple(sorted(e)) for e in get_redundant_edges_from_assembly_points(e=(u, v), points_by_edges=assembly_points_by_edges,
                                                                                      processed_points=already_processed_assembly_points)}
            redundant_edges.add(tuple(sorted((u, v))))
            redundant_edges.update(re1)
            redundant_edges.update(re2)
            redundant_edges.update(re3)
        for u, v in redundant_edges:
            to_discard = (u, v, merged_scaffold_assembly_edges_graph.graph[u][v]['weight'])
            assembly_edges.discard(to_discard)
    # checking that we didn't screw up :)
    for vertex in cover_graph.nodes():
        if cover_graph.degree(vertex) > 2:
            print(vertex)
            print(cover_graph.edges(nbunch=vertex))
            exit(1)
    return cover_graph


####################################################################
####################################################################
#                                                                  #
# Maximal matching strategy for combining scaffold assemblies      #
#                                                                  #
####################################################################

def maximal_matching(assembly_points_by_sources, acyclic=True, min_cw=0.0):
    assembly_points_by_sources = [ap for ap_list in assembly_points_by_sources.values() for ap in ap_list]
    scaffold_edges = get_scaffold_edges(assembly_points=assembly_points_by_sources)
    unoriented_assembly_points = get_un_oriented_assembly_points(assembly_points=assembly_points_by_sources)

    assembly_edges_graph = MergedScaffoldAssemblyGraph()
    for ap in assembly_points_by_sources:
        for (u, v, weight) in ap.get_edges(sort=True, weight=True):
            assembly_edges_graph.add_edge(u, v, weight=weight)
    assembly_edges_graph.remove_edges_with_low_cw(cw_threshold=min_cw)
    matching = networkx.max_weight_matching(G=assembly_edges_graph.graph)
    cover_graph = networkx.Graph()
    cover_graph.add_edges_from(scaffold_edges)
    edges = get_edges_from_matching(matching)
    for u, v in edges:
        cover_graph.add_edge(u, v, weight=assembly_edges_graph.graph[u][v]['weight'])
    # sanity check
    for vertex in cover_graph.nodes():
        assert cover_graph.degree(vertex) <= 2
    # checking for any issues with unoriented assembly points
    for assembly_point in unoriented_assembly_points:
        participating_edges = []
        for (u, v) in assembly_point.get_edges():
            if cover_graph.has_edge(u=u, v=v):
                participating_edges.append((u, v))
        assert len(participating_edges) <= 2  # this is just a sanity check
        if len(participating_edges) == 2:
            cover_graph.remove_edge(participating_edges[0][0], participating_edges[0][1])
    if acyclic:
        edges_to_delete = []
        for cc in networkx.connected_component_subgraphs(G=cover_graph, copy=True):
            if networkx.number_of_nodes(cc) == networkx.number_of_edges(cc):
                assembly_edges = filter(lambda entry: "weight" in entry[2], cc.edges(data=True))
                edges_to_delete.append(min(assembly_edges, key=lambda entry: entry[2]["weight"]))
        for u, v, data in edges_to_delete:
            cover_graph.remove_edge(u, v)
    return cover_graph


def get_edges_from_matching(matching):
    seen = set()
    edges = []
    for key, value in matching.items():
        if key not in seen and value not in seen:
            edges.append((key, value))
            seen.add(key)
            seen.add(value)
    return edges


strategies_bindings = {
    MergingStrategies.greedy_merging.value: merge_greedily,
    MergingStrategies.maximal_matching.value: maximal_matching
}


def update_assembly_points_with_merged_assembly(original_assembly_points_by_ids, merged_assembly_points_by_ids, merged_assembly_graph):
    for ap in merged_assembly_points_by_ids.values():
        for u, v in ap.get_edges():
            if merged_assembly_graph.has_edge(u=u, v=v):
                par_or_1 = "+" if u.endswith("h") else "-"
                par_or_2 = "+" if v.endswith("t") else "-"
                forward = u[:-1] == ap.seq1
                if not forward:
                    par_or_1, par_or_2 = inverse_orientation(par_or_2), inverse_orientation(par_or_1)
                ap.seq1_par_or = par_or_1
                ap.seq2_par_or = par_or_2
                ap.participates_in_merged = True
                for child_id in ap.children_ids:
                    child = original_assembly_points_by_ids[child_id]
                    child.participates_in_merged = True
                    child.seq1_par_or = par_or_1
                    child.seq2_par_or = par_or_2
                break


def update_gap_sizes_in_merged_assembly(original_assembly_points_by_ids, merged_assembly_points_by_ids):
    aps_in_merged_assembly = [ap for ap in merged_assembly_points_by_ids.values() if ap.participates_in_merged]
    for ap in aps_in_merged_assembly:
        children_aps = [original_assembly_points_by_ids[c_id] for c_id in ap.children_ids]
        cumulative_gap_size = 0.0
        gap_size_sources_cnt = 0.0
        for c_ap in children_aps:
            if isinstance(c_ap.gap_size, numbers.Number):
                gap_size_sources_cnt += 1
                cumulative_gap_size += c_ap.gap_size
        if gap_size_sources_cnt == 0:
            inferred_gap_size = "?"
        else:
            inferred_gap_size = cumulative_gap_size / gap_size_sources_cnt
        ap.gap_size = inferred_gap_size



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import datetime
import itertools
import logging
import os
import shutil
import sys
from collections import defaultdict

import configargparse
import six
from jinja2 import FileSystemLoader
from jinja2.environment import Environment

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import camsa
from camsa.core import io as camsa_io
from camsa.core import merging
from camsa.core.reference_analysis import analyze_and_update_assembly_points_based_on_reference
from camsa.core.comparative_analysis import compute_and_update_assembly_points_conflicts
from camsa.core.data_structures import Assembly, assign_ids_to_assembly_points, merge_assembly_points, assign_parents_to_children, to_json, Sequence
from camsa.core.merging import MergingStrategies, update_assembly_points_with_merged_assembly, update_gap_sizes_in_merged_assembly

if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="CAMSA is a tool for Comparative Analysis and Merging of Scaffold Assemblies",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"

    parser = configargparse.ArgParser(description=full_description,
                                      formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "run_camsa.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])
    parser.add_argument("points", nargs="+",
                        help="A list of input files, representing a standard CAMSA format for assembly points.")
    parser.add_argument("--seqi", default=None,
                        help="A file with sequences' information about fragments, involved in input assembly points")
    parser.add_argument("--seqi-delimiter", type=str, default="\t",
                        help="A delimiter character for the file containing sequences' information.\nDEFAULT: \\t")
    parser.add_argument("--seqi-ensure-all", action="store_true", dest="lengths_ensure", default=False,
                        help="Ensures, that either all, or non of the sequences participating in assembly points have information provided for them.\nDEFAULT: False")
    parser.add_argument("-c", "--config", is_config_file=True,
                        help="Config file path with settings for CAMSA to run with.\nOverwrites the default CAMSA configuration file.\nValues in config file can be overwritten by command line arguments.")
    parser.add_argument("--c-cw-exact", type=float,
                        help="A confidence weight value assigned to oriented assembly points and respective exact assembly edges,\nin case \"?\" is specified as the respective assembly point confidence weight.\nDEFAULT: 1.0")
    parser.add_argument("--c-cw-candidate", type=float,
                        help="A confidence weight value assigned to semi/un-oriented assembly points and respective candidate assembly edges,\nin case \"?\" is specified as the respective assembly point confidence weight.\nDEFAULT: 0.75")
    # parser.add_argument("--c-comparative-disable", action="store_false", dest="comparative", default=True,
    #                     help="")
    parser.add_argument("--ref-disable", action="store_false", dest="reference", default=True,
                        help="")
    parser.add_argument("--ref", type=str, default="", dest="reference_name",
                        help="")
    # parser.add_argument("--c-merging-disable", action="store_false", dest="merging", default=True,
    #                     help="")
    parser.add_argument("--c-merging-cw-min", type=float,
                        help="A threshold for the minimum cumulative confidence weight for merged assembly edges in MSAG.\nEdges with confidence weight below are not considered in the \"merged\" assembly construction.\nDEFAULT: 0.0")
    parser.add_argument("--c-merging-strategy", choices=[MergingStrategies.greedy_merging.value, MergingStrategies.maximal_matching.value],
                        default=MergingStrategies.maximal_matching.value,
                        help="A strategy to produced a merged assembly from the given ones.\nDEFAULT: maximal-matching")
    parser.add_argument("--c-merging-cycles", dest="allow_cycles", action="store_true", default=False,
                        help="Whether to allow cycles in the produced merged assembly.\nDEFAULT: False")
    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("--i-delimiter", default="\t", type=str,
                        help="String used as a delimiter in the input files with CAMSA assembly points")
    parser.add_argument("-o", "--o-dir",
                        help="A directory, where CAMSA will store all of the produced output (report, assets, etc).\nDEFAULT: camsa_{date}")
    # parser.add_argument("--o-interactive-disable", action="store_false", dest="o_interactive", default=True,
    #                     help="")
    parser.add_argument("--o-merged-format", type=str,
                        help="The CAMSA-out formatting for the merged scaffold assemblies in a form of CAMSA points.")
    parser.add_argument("--o-subgroups-format", type=str,
                        help="The CAMSA-out formatting for the merged scaffold assemblies in a form of CAMSA points.")
    parser.add_argument("--o-collapsed-format", type=str,
                        help="The CAMSA-out formatting for the collapsed assembly points and their computed conflicts.")
    parser.add_argument("--o-original-format", type=str,
                        help="The CAMSA-out formatting for the non-collapsed assembly points and their computed conflicts.")
    parser.add_argument("--c-logging-level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for CAMSA.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    args = parser.parse_args()

    start_time = datetime.datetime.now()

    #######################################
    #           logging setup             #
    #######################################
    if args.o_dir is None:
        args.o_dir = os.path.join(os.getcwd(), "camsa_{date}".format(
            date=datetime.datetime.now().strftime("%b_%d_%Y__%H_%M")))

    args.output_dir = os.path.abspath(os.path.expanduser(args.o_dir))
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    logger = logging.getLogger("CAMSA.main")
    ch = logging.StreamHandler()
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    fh = logging.FileHandler(os.path.join(args.output_dir, "camsa.log"), "wt")
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    fh.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the analysis")

    #######################################
    #           input stage               #
    #######################################
    logger.info("Processing input")

    # key is the origin of assembly point; values is the list of all points from that source
    logger.info("Reading assembly points")
    assembly_points_by_sources = camsa_io.read_assembly_points_from_input_sources(sources=args.points,
                                                                                  default_cw_eae=args.c_cw_exact,
                                                                                  default_cw_cae=args.c_cw_candidate,
                                                                                  delimiter=args.i_delimiter)
    or_seqi = defaultdict(list)
    if args.seqi is not None:
        args.seqi = os.path.abspath(os.path.expanduser(args.seqi))
        logger.info("Reading sequences' info")
        with open(args.seqi, "rt") as source:
            camsa_io.read_seqi_from_input_sources(source=source, delimiter=args.seqi_delimiter, destination=or_seqi)
    seqi = {}
    for seq_name in list(or_seqi.keys()):
        entry_list = or_seqi[seq_name]
        seqi[seq_name] = entry_list[0]
    if len(seqi) > 0:
        sequences_ids_from_assembly_points = set()
        for source, aps in assembly_points_by_sources.items():
            for ap in aps:
                sequences_ids_from_assembly_points.add(ap.seq1)
                sequences_ids_from_assembly_points.add(ap.seq2)
        sequences_ids_from_lengths = set(seqi.keys())
        difference = sequences_ids_from_assembly_points - sequences_ids_from_lengths
        if len(difference) > 0:
            logger.warning("Some sequences, that participate in assembly points do not have lengths associated with them,"
                           "while others do.")
            logger.warning("Problematic sequences: {seqs}".format(seqs=",".join(sorted(difference))))
            if args.seqi_ensure:
                logger.error("A flag \"--seqi-ensure-all\" was set, so the program terminates")
            else:
                logger.warning("Assigning lengths of -1 to all problematic sequences")
                for seq_id in difference:
                    seqi[seq_id] = Sequence(name=seq_id, length=-1)

    #######################################
    #       extracting reference          #
    #######################################
    reference_assembly = None
    if args.reference and args.reference_name != "":
        try:
            reference_aps = assembly_points_by_sources.pop(args.reference_name)
            reference_assembly = Assembly(name=args.reference_name, aps=reference_aps)
        except KeyError:
            logger.critical("Supplied reference \"{reference_name}\" was not found among assembly sources [{avail_sources}]".format(reference_name=args.reference_name, avail_sources=",".join(assembly_points_by_sources.keys())))
            exit(1)

    id_generator = itertools.count()
    original_assembly_points_by_ids = {}
    for assembly_points in assembly_points_by_sources.values():
        original_assembly_points_by_ids.update(assign_ids_to_assembly_points(assembly_points=assembly_points, id_prefix="or_",
                                                                             id_generator=id_generator))

    #######################################
    #       assembly points merging       #
    #######################################
    logger.info("Merging assembly points from different sources into a set of unique ones.")
    merged_assembly_points = merge_assembly_points(assembly_points_by_source=assembly_points_by_sources)
    merged_assembly_points_by_ids = assign_ids_to_assembly_points(assembly_points=merged_assembly_points, id_prefix="m_")
    assign_parents_to_children(children_assembly_points_by_ids=original_assembly_points_by_ids,
                               parent_assembly_points_by_ids=merged_assembly_points_by_ids)

    #######################################
    #       assemblies subgroups          #
    #######################################
    logger.info("Processing assemblies' subgroups")
    tmp_individual_assemblies = defaultdict(list)
    for ap in merged_assembly_points:
        for source_name in ap.sources:
            tmp_individual_assemblies[source_name].append(ap)
    individual_assemblies = [Assembly(name=name, aps=aps) for name, aps in tmp_individual_assemblies.items()]

    grouped_assemblies = []
    for i in range(1, len(individual_assemblies) + 1):
        for assembly_combination in itertools.combinations(sorted([a.name for a in individual_assemblies]), i):
            assembly = Assembly(name=assembly_combination, aps=[])
            for ap in merged_assembly_points:
                if len(ap.sources) == len(assembly_combination) and all(map(lambda entry: entry in ap.sources, assembly_combination)):
                    assembly.aps.append(ap)
            grouped_assemblies.append(assembly)
    grouped_assemblies = list(filter(lambda a: len(a.aps) > 0, sorted(grouped_assemblies, key=lambda entry: len(entry.aps), reverse=True)))

    #######################################
    #        reference   analysis         #
    #######################################
    # for assembly in individual_assemblies:
    #     analyze_and_update_assembly_points_based_on_reference(assembly=assembly, ref_assembly=reference_assembly)

    #######################################
    #        comparative analysis         #
    #######################################
    logger.info("Computing assembly points conflicts")
    compute_and_update_assembly_points_conflicts(assembly_points_by_ids=merged_assembly_points_by_ids)

    #######################################
    #       merging assemblies            #
    #######################################
    logger.info("Obtaining a merged assembly, using {strategy} strategy".format(strategy=args.c_merging_strategy))
    merged_assembly_graph = merging.strategies_bindings[args.c_merging_strategy](assembly_points_by_sources=assembly_points_by_sources,
                                                                                 acyclic=not args.allow_cycles,
                                                                                 min_cw=args.c_merging_cw_min)
    update_assembly_points_with_merged_assembly(original_assembly_points_by_ids=original_assembly_points_by_ids,
                                                merged_assembly_points_by_ids=merged_assembly_points_by_ids,
                                                merged_assembly_graph=merged_assembly_graph)
    update_gap_sizes_in_merged_assembly(original_assembly_points_by_ids=original_assembly_points_by_ids,
                                        merged_assembly_points_by_ids=merged_assembly_points_by_ids)

    #######################################
    #           output stage              #
    #######################################
    logger.info("Preparing output")

    # copying assets required for the HTML report
    libs_report_dir = os.path.join(args.output_dir, "libs")
    camsa_io.remove_dir(dir_path=libs_report_dir)
    shutil.copytree(src=os.path.join(camsa.root_dir, "libs"), dst=libs_report_dir)
    output_html_report_file_name = os.path.join(args.output_dir, "report.html")

    # "input" subdir of the report
    # will contain a configuration as well as assembly points files
    input_report_dir = os.path.join(args.output_dir, "input")
    camsa_io.remove_dir(dir_path=input_report_dir)
    os.makedirs(input_report_dir)
    input_report_config_path = os.path.join(input_report_dir, "camsa_config.txt")
    with open(input_report_config_path, "wt") as destination:
        print("# NOTE: this is not a valid config, but rather a summary of the utilized options", file=destination)
        print(parser.format_values(), file=destination)
    for pairs_path in args.points:
        full_path = os.path.abspath(os.path.expanduser(pairs_path))
        base_name = os.path.basename(full_path)
        shutil.copyfile(src=full_path, dst=os.path.join(input_report_dir, base_name))

    # "merged" subdir of the report
    # will contain assembly points, that constitute the merged assembly
    merged_report_dir = os.path.join(args.output_dir, "merged")
    camsa_io.remove_dir(dir_path=merged_report_dir)
    os.makedirs(merged_report_dir)
    merged_report_points_path = os.path.join(merged_report_dir, "merged.camsa.points")
    with open(merged_report_points_path, "wt") as destination:
        camsa_io.write_assembly_points(destination=destination,
                                       assembly_points=[ap for ap in merged_assembly_points if ap.participates_in_merged],
                                       output_setup=args.o_merged_format)

    # "comparative" subdir of the report
    # will contain assembly points divided into subgroups based in the agreement in input assemblies
    comparative_report_dir = os.path.join(args.output_dir, "comparative")
    camsa_io.remove_dir(dir_path=comparative_report_dir)
    subgroups_report_dir = os.path.join(comparative_report_dir, "subgroups")
    camsa_io.remove_dir(comparative_report_dir)
    os.makedirs(comparative_report_dir)
    os.makedirs(subgroups_report_dir)
    for group in grouped_assemblies:
        comparative_report_group_points_path = os.path.join(subgroups_report_dir, "{group_name}.camsa.points".format(group_name=".".join(group.name)))
        with open(comparative_report_group_points_path, "wt") as destination:
            camsa_io.write_assembly_points(assembly_points=group.aps,
                                           destination=destination,
                                           output_setup=args.o_subgroups_format)

    original_points_path = os.path.join(comparative_report_dir, "original.camsa.points")
    with open(original_points_path, "wt") as destination:
        camsa_io.write_assembly_points(assembly_points=original_assembly_points_by_ids.values(),
                                       destination=destination,
                                       output_setup=args.o_original_format)

    collapsed_points_path = os.path.join(comparative_report_dir, "collapsed.camsa.points")
    with open(collapsed_points_path, "wt") as destination:
        camsa_io.write_assembly_points(destination=destination,
                                       assembly_points=merged_assembly_points,
                                       output_setup=args.o_collapsed_format)
    env = Environment()
    env.filters['tojson'] = to_json

    individual_assemblies.sort(key=lambda it: it.name.lower())
    assemblies_to_ids = {assembly.name: "A" + str(cnt) for cnt, assembly in enumerate(individual_assemblies, start=1)}
    sources = [assembly.name for assembly in individual_assemblies]

    assemblies_to_colors = {assemblies_to_ids[source]: color for source, color in zip(sources, ['red', 'blue', 'green', 'purple',
                                                                                                'orange', 'pink', 'brown', 'navy', 'steelblue'])}

    with open(output_html_report_file_name, "wt") as dest:
        env.loader = FileSystemLoader(os.path.join(camsa.root_dir, "html"))
        template = env.get_template("base_template.html")
        six.print_(template.render(
            data={
                "assemblies": individual_assemblies,
                "assemblies_intersections": [],
                "assemblies_conflicts": [],
                "graph_compiled": False,
                "aps": merged_assembly_points,
                "assemblies_to_ids": assemblies_to_ids,
                "assemblies_to_colors": assemblies_to_colors,
                "grouped_assemblies": grouped_assemblies,
                "fragments": {
                    "seqi": seqi,
                    'max_length': max([seq.length for seq in seqi.values()]) if len(seqi) > 0 else -1,
                    'min_length': min([seq.length for seq in seqi.values()]) if len(seqi) > 0 else -1,
                }
            },
            settings={
                "cytoscape": {
                    "draw_timeout": 300000
                }
            },
            meta={
                "camsa": {
                    "version": camsa.VERSION
                },
                "date": start_time
            }), file=dest)
    logger.info("CAMSA report is written to \"{output_report_file}\"".format(output_report_file=output_html_report_file_name))
    logger.info("Finished Comparative Analysis and Merging of input assemblies.")
    end_time = datetime.datetime.now()
    logger.info("Elapsed time: {el_time}".format(el_time=str(end_time - start_time)))
    logger.info("Thank you for using CAMSA!")



# -*- coding: utf-8 -*-
import networkx as nx
from camsa.core.data_structures import ScaffoldAssemblyGraph


def analyze_and_update_assembly_points_based_on_reference(assembly, ref_assembly):
    reference_graph = ScaffoldAssemblyGraph.from_assembly_points(assembly_points=ref_assembly.aps, reference=True)
    ref_all_pairs_paths = nx.all_pairs_shortest_path(G=reference_graph.graph)


def has_unoriented_assembly_points(assembly):
    return not all(map(lambda ap: ap.is_oriented, assembly.aps))






# -*- coding: utf-8 -*-

def get_alignments(iseq, jseq, backtracking, end_cell):
    iseq_r = []
    jseq_r = []
    current_cell_v = backtracking[end_cell[0]][end_cell[1]]
    current_cell_index = end_cell
    while current_cell_v != "s":
        if current_cell_v == "u":
            jseq_r.append("-")
            iseq_r.append(iseq[current_cell_index[0] - 1])
            current_cell_index = (current_cell_index[0] - 1, current_cell_index[1])
        elif current_cell_v == "l":
            jseq_r.append("-")
            jseq_r.append(jseq_r[current_cell_index[1] - 1])
            current_cell_index = (current_cell_index[0], current_cell_index[1] - 1)
        else:
            iseq_r.append(iseq[current_cell_index[0] - 1])
            jseq_r.append(jseq[current_cell_index[1] - 1])
            current_cell_index = (current_cell_index[0] - 1, current_cell_index[1] - 1)
        current_cell_v = backtracking[current_cell_index[0]][current_cell_index[1]]
    return "".join(reversed(iseq_r)), "".join(reversed(jseq_r))


def get_consensus_sequences(aligned_seqs):
    result = []
    for fl, sl in zip(*aligned_seqs):
        if fl == "-":
            candidate = sl
        elif sl == "-":
            candidate = fl
        elif fl == "N":
            candidate = sl
        elif sl == "N":
            candidate = fl
        elif sl == fl:
            candidate = sl
        else:
            candidate = "N"
        result.append(candidate)
    return "".join(result)


def substitute_score(letter1, letter2):
    if letter1 == "N" or letter2 == "N" or letter1 == letter2:
        return 0
    return 1


def bounded_alignment(seq1, seq2, max_edit_distance=-1):
    if max_edit_distance == -1:
        max_edit_distance = max(len(seq1), len(seq2))
    i_range = len(seq1) + 1
    j_range = len(seq2) + 1
    M = [[0] * j_range for _ in range(i_range)]
    Mbt = [["s" for _ in range(j_range)] for _ in range(i_range)]
    for j in range(1, j_range):
        M[0][j] = j
        Mbt[0][j] = "l"
    for i in range(1, i_range):
        M[i][0] = i
        Mbt[i][0] = "u"
    for i in range(1, i_range):
        for j in range(1, j_range):
            if abs(i - j) > max_edit_distance:
                continue
            options = [(M[i - 1][j - 1] + substitute_score(seq1[i - 1], seq2[j - 1]), "d")]
            if abs(i - j) < max_edit_distance:
                options.append((M[i - 1][j] + 1, "u"))
                options.append((M[i][j - 1] + 1, "l"))
            else:
                if i > j:
                    options.append((M[i - 1][j] + 1, "u"))
                else:
                    options.append((M[i][j - 1] + 1, "l"))
            result = min(options, key=lambda entry: entry[0])
            M[i][j] = result[0]
            Mbt[i][j] = result[1]
    if M[-1][-1] > max_edit_distance:
        return AlignmentResult(seq1=seq1, seq2=seq2, consensus=None, al_seq1=None, al_seq2=None, edit_distance=M[-1][-1])
    aligned_seqs = get_alignments(iseq=seq1, jseq=seq2, backtracking=Mbt, end_cell=(i_range - 1, j_range - 1))
    consensus_seq = get_consensus_sequences(aligned_seqs=aligned_seqs)
    return AlignmentResult(seq1=seq1, seq2=seq2, consensus=consensus_seq, al_seq1=aligned_seqs[0], al_seq2=aligned_seqs[1], edit_distance=M[-1][-1])


class AlignmentResult(object):
    def __init__(self, seq1, seq2, consensus, al_seq1, al_seq2, edit_distance):
        self.seq1 = seq1
        self.seq2 = seq2
        self.al_seq1 = al_seq1
        self.al_seq2 = al_seq2
        self.consensus = consensus
        self.edit_distance = edit_distance



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division

import datetime
import logging
import numbers
import os
import sys
from collections import defaultdict, Counter

import Bio
import configargparse
import itertools
import networkx
from Bio import SeqIO
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
from camsa.core.io import read_pairs, read_seqi_from_input_sources
from camsa.core.data_structures import get_scaffold_edges, Sequence
from camsa.utils.fasta.data_structures import IntraGapFilling, FlankingGapFilling


def get_scaffold_name_from_vertex(v):
    return v[:-1]


def reverse_or(orientaiton):
    return "+" if orientaiton == "-" else "-"


def reverse_ap(f1, f1_or, f2, f2_or):
    return f2, reverse_or(orientaiton=f2_or), f1, reverse_or(orientaiton=f1_or)


def get_assembly_edge(graph):
    for edge in graph.edges():
        v1, v2 = edge
        if v1[:-1] != v2[:-1]:
            return edge
    return None, None


def get_sequence_of_fragments_from_path(path, assembly_points_by_edges):
    path, path_type = path
    if len(path) < 2:
        logger.error("A sequence resembling an assembly points, that contains less than two scaffolds. Something went wrong.")
        exit(1)
    result = []
    for frag_extremity_v1, frag_extremity_v2 in zip(path[1::2], path[2::2]):
        f1_or = "+" if frag_extremity_v1.endswith("h") else "-"
        f2_or = "-" if frag_extremity_v2.endswith("h") else "+"
        ap = assembly_points_by_edges[tuple(sorted([frag_extremity_v1, frag_extremity_v2]))]
        gap_size = ap.gap_size
        result.append((get_scaffold_name_from_vertex(v=frag_extremity_v1), f1_or,
                       get_scaffold_name_from_vertex(v=frag_extremity_v2), f2_or,
                       gap_size))
    return result


def get_fasta_for_meta_seq(meta_seq, orientation, fasta_by_ids):
    seq = fasta_by_ids[meta_seq.parent_seq_id].seq[meta_seq.start:meta_seq.end]
    seq = seq if meta_seq.strand == "+" else seq.reverse_complement()
    seq = seq if orientation == "+" else seq.reverse_complement()
    return seq


def get_meta_flanking_by_meta_seq_name(meta_seq_name, orientation, meta_sequences_by_name, meta_sequences_by_parent_ids, fasta_by_ids, extension_type):
    """

    :param meta_seq_name:
    :param orientation:
    :param meta_sequences_by_name:
    :param meta_sequences_by_parent_ids:
    :param fasta_by_ids:
    :param extension_type: "before" / "after"
    :return:
    """
    meta_sequences = meta_sequences_by_name[meta_seq_name]
    result = []
    for meta_sequence in meta_sequences:
        flanking_meta_seq = get_meta_flanking_by_meta_seq(meta_sequence=meta_sequence, extension_type=extension_type, fasta_by_ids=fasta_by_ids, meta_sequences_by_parent_ids=meta_sequences_by_parent_ids, orientation=orientation)
        if flanking_meta_seq is not None and flanking_meta_seq.length > 0:
            result.append(flanking_meta_seq)
    return result


def get_meta_flanking_by_meta_seq(extension_type, fasta_by_ids, meta_sequence, meta_sequences_by_parent_ids, orientation):
    if meta_sequence.name == meta_sequences_by_parent_ids[meta_sequence.parent_seq_id][0].name and meta_sequence.start > 0:
        if extension_type == "before" and meta_sequence.strand == orientation:
            extension_meta_seq = Sequence(name="extension",
                                          parent_seq_id=meta_sequence.parent_seq_id,
                                          start=0,
                                          end=meta_sequence.start,
                                          strand="+",
                                          seq_group_id=meta_sequence.seq_group_id)

            return extension_meta_seq
        elif extension_type == "after" and meta_sequence.strand != orientation:
            extension_meta_seq = Sequence(name="extension",
                                          parent_seq_id=meta_sequence.parent_seq_id,
                                          start=0,
                                          end=meta_sequence.start,
                                          strand="-",
                                          seq_group_id=meta_sequence.seq_group_id)

            return extension_meta_seq
    elif meta_sequence.name == meta_seqs_by_parent_ids[meta_sequence.parent_seq_id][-1].name and meta_sequence.end < len(fasta_by_ids[meta_sequence.parent_seq_id]):
        if extension_type == "before" and meta_sequence.strand != orientation:
            extension_meta_seq = Sequence(name="extension",
                                          parent_seq_id=meta_sequence.parent_seq_id,
                                          start=meta_sequence.end,
                                          end=len(fasta_by_ids[meta_sequence.parent_seq_id]),
                                          strand="-",
                                          seq_group_id=meta_sequence.seq_group_id)
            return extension_meta_seq
        elif extension_type == "after" and meta_sequence.strand == orientation:
            extension_meta_seq = Sequence(name="extension",
                                          parent_seq_id=meta_sequence.parent_seq_id,
                                          start=meta_sequence.end,
                                          end=len(fasta_by_ids[meta_sequence.parent_seq_id]),
                                          strand="+",
                                          seq_group_id=meta_sequence.seq_group_id)
            return extension_meta_seq


def collinearity(meta_seq1, meta_seq2, orientation1, orientation2):
    if meta_seq1.strand == orientation1 and meta_seq2.strand == orientation2 and meta_seq1.end <= meta_seq2.start:
        return 1
    elif meta_seq1.strand == reverse_or(orientaiton=orientation1) and meta_seq2.strand == reverse_or(orientaiton=orientation2) and meta_seq2.end <= meta_seq1.start:
        return -1
    return 0


def get_dummy_gap_filling(gap_size, sep, default_gap_size):
    if gap_size == "?":
        gap_size = default_gap_size
    result = Seq(sep * int(gap_size))
    return result


if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting CAMSA formatted scaffolding results into FASTA files.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "utils", "fasta", "camsa_points2fasta.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])
    parser.add_argument("-c", "--config", is_config_file=True, help="Config file overwriting some of the default settings as well as any flag starting with \"--\".")

    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("--fasta", type=configargparse.FileType("rt"), required=True,
                        help="A stream of fasta formatted sequences of fragments, that participate in the scaffold assembly represented in form of CAMSA points")
    parser.add_argument("--points", type=configargparse.FileType("rt"), required=True,
                        help="A stream of CAMSA formatted assembly points, representing a scaffold assembly, that is converted into FASTA formatted sequences")
    parser.add_argument("--seqi", default="", type=str,
                        help="A stream of CAMSA formatted information about fragments, that participate in the reported assembly points.\nDEFAULT: dummy seqi records are created for each fasta record, if non provided.")
    parser.add_argument("--seqi-delimiter", default="\t", type=str,
                        help="A delimiter character for the file containing sequences' information.\nDEFAULT: \\t")
    parser.add_argument("--genomes-order", default="", type=str,
                        help="Order in which genomes sequences (i.e., seq_group_id column in *.seqi file) shall be for resulting sequence reconstruction.\nDEFAULT: sorted order of all seq_group_id column values. Seqi entries without such attribute are assigned the \"Default\" seq_group and used last.")
    parser.add_argument("--extend-ends", action="store_true", dest="extend_ends",
                        help="Flag specifying whether to extend the extremities in the reconstructed assembly, when possible, using flanking extremities in the sequences, from which fragments come.\nDEFAULT: True", default=True)
    parser.add_argument("--no-extend-ends", action="store_false", dest="extend_ends",
                        help="Flag specifying whether to extend the extremities in the reconstructed assembly, when possible, using flanking extremities in the sequences, from which fragments come.\nDEFAULT: True", default=True)
    parser.add_argument("--fill-gaps", action="store_true", dest="fill_gaps",
                        help="Flag specifying whether to fill gaps in the reconstructed assembly using information in the sequences, from which fragments participating in assembly points come.\nDEFAULT: True", default=True)
    parser.add_argument("--no-fill-gaps", action="store_false", dest="fill_gaps",
                        help="Flag specifying whether to fill gaps in the reconstructed assembly using information in the sequences, from which fragments participating in assembly points come.\nDEFAULT: True", default=True)
    parser.add_argument("--fill-gaps-unknown", action="store_true",
                        help="A flag specifying whether to fill gaps with unknown gap size or nor.\nDEFAULT: False (i.e., Ns will be inserted)")
    parser.add_argument("--gap-diff-threshold-per", default=10.0, type=float,
                        help="A percentage wise difference (i.e., in [0, 100]) specifying how much the length of the filling can differ from the given gap size.\nDEFAULT: 10")
    parser.add_argument("--gap-diff-threshold-bp", default=1000, type=float,
                        help="A bp wise difference (i.e., [0,..]) specifying how much the length of the filling can differ from the given gap size.\nDEFAULT: 10")
    parser.add_argument("--gap-diff-threshold-max", action="store_true", dest="gap_diff_max_over_min", default=False,
                        help="A flag specifying whether to take maximum possible difference (i.e., between percentage wise and bp wise) or a minimum one.\nDEFAULT: False (i.e., minimum, more conservative, difference is taken as a threshold)")
    parser.add_argument("--gap-fill-no-intra-first", action="store_false", dest="gap_fill_intra_first", default=True,
                        help="A flag specifying whether gaps shall be fist filled with intra-fragment sequence, rather than using two flanking, possible overlapping sequences.\nDEFAULT: True")
    parser.add_argument("--gap-fill-no-inter", action="store_false", dest="gap_fill_inter", default=True,
                        help="A flag specifying whether it is allowed to fill gaps with pairs flanking sequences or not.")
    # parser.add_argument("--allow-singletons", action="store_true", dest="allow_singletons", default=False,
    #                     help="Whether to include scaffolds, that were not mentioned in the CAMSA formatted assembly points or in there \nDEFAULT: False")
    parser.add_argument("--c-sep", type=str,
                        help="A symbol, that is used to indicate gaps between scaffolds in the translated assemblies\nDEFAULT: N")
    parser.add_argument("--c-sep-length", type=int,
                        help="A default length, that is used for the gap size between scaffolds in the translated assemblies. Used in case, when gap-size column has \"?\" value\nDEFAULT: 20")
    parser.add_argument("--scaffold-name-template", type=str,
                        help="Python string template for the scaffold ids, in the produced FASTA formatted sequences. \"cnt\" attribute can be utilized\nDEFAULT: scaffold_{cnt}")
    parser.add_argument("-o", "--output", type=configargparse.FileType("wt"), default=sys.stdout,
                        help="A stream to which the FASTA formatted converted sequence, representing the CAMSA formatted scaffold assembly, is output\nDEFAULT: stdout")

    parser.add_argument("--c-logging-level", dest="logging_level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    args = parser.parse_args()
    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.camsa_points2fasta")
    ch = logging.StreamHandler()
    ch.setLevel(args.logging_level)
    logger.setLevel(args.logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    logger.info("Reading assembly points")
    assembly_points_by_sources = defaultdict(list)
    read_pairs(source=args.points, destination=assembly_points_by_sources)
    assembly_points_by_sources = [ap for ap_list in assembly_points_by_sources.values() for ap in ap_list]
    logger.info("A total of {ap_cnt} assembly points was obtained".format(ap_cnt=len(assembly_points_by_sources)))

    assembly_graph = networkx.Graph()

    scaffold_edges = get_scaffold_edges(assembly_points=assembly_points_by_sources)
    assembly_graph.add_edges_from(ebunch=scaffold_edges)
    assembly_points_by_edges = {}

    for ap in assembly_points_by_sources:
        for (u, v, weight) in ap.get_edges(sort=True, weight=True):
            assembly_graph.add_edge(u, v)
            assembly_points_by_edges[tuple(sorted([u, v]))] = ap

    logger.debug("Checking that there are no in(semi)conflicting assembly points")
    for vertex in assembly_graph.nodes():
        degree = assembly_graph.degree(nbunch=vertex)
        if degree > 2:
            scaffold_name = get_scaffold_name_from_vertex(v=vertex)
            logger.error("Supplied assembly contained a conflict.")
            logger.error("Scaffold {scaffold_name} by its extremity {extremity_name} is reported as adjacent to more than one other scaffold's extremity"
                         "".format(scaffold_name=scaffold_name, extremity_name=vertex))
            exit(1)
    logger.debug("All clear, no (semi)conflicts, we can proceed")

    logger.info("Processing assemblies constructed from obtained assembly points")
    logger.debug("Extracting paths from assembly graph")
    paths = []
    for cc in networkx.connected_component_subgraphs(G=assembly_graph):
        origins = [v for v in cc.nodes() if cc.degree(v) == 1]
        if len(origins) == 2:
            path = networkx.shortest_path(G=cc, source=origins[0], target=origins[1])
            logger.debug("Extracted a linear scaffold of length {scaffold_length}, staring with {s_v} and ending with {e_v}"
                         "".format(scaffold_length=int(len(path) / 2),
                                   s_v=get_scaffold_name_from_vertex(v=origins[0]),
                                   e_v=get_scaffold_name_from_vertex(v=origins[1])))
            paths.append((path, "l"))
        if len(origins) == 1:
            logger.error("Something is wrong with the assembly graph. We have a connected component with a single vertex of degree 1.")
            exit(1)
        if len(origins) == 0:
            logger.debug("Encountered a circular chromosome. Splitting it at random assembly point")
            assembly_edge = get_assembly_edge(graph=cc)
            if assembly_edge[0] is None or assembly_edge[1] is None:
                logger.error("Something is wrong with the assembly graph. Couldn't find a scaffold edge in a circular scaffold.")
                exit(1)
            cc.remove_edge(u=assembly_edge[0], v=assembly_edge[1])
            path = networkx.shortest_path(G=cc, source=assembly_edge[0], target=assembly_edge[1])
            paths.append((path, "c"))
    logger.debug("Total number of extracted paths is {path_cnt}".format(path_cnt=len(paths)))
    logger.debug("Out of which {linear_cnt} are linear, and {circular_cnt} are circular"
                 "".format(linear_cnt=len([p for p in paths if p[1] == "l"]),
                           circular_cnt=len([p for p in paths if p[1] == "c"])))
    fragments = [get_sequence_of_fragments_from_path(path=p, assembly_points_by_edges=assembly_points_by_edges) for p in paths]
    logger.info("Total number of {scaffold_cnt} scaffolds was obtained from observed assembly points".format(scaffold_cnt=len(fragments)))

    logger.info("Reading fasta of contigs/scaffolds from {file}".format(file=args.fasta))
    frag_fasta_by_id = {}
    s_cnt = 0
    for record in SeqIO.parse(args.fasta, "fasta"):
        frag_fasta_by_id[record.id] = record
        s_cnt += 1
    logger.debug("Obtained {s_cnt} fasta fragments".format(s_cnt=s_cnt))

    sequences_by_ids = defaultdict(list)
    if args.seqi != "":
        logger.debug("Obtaining \"meta\"-sequence info for the fragments, that can be involved in the assembly points, from {file}".format(file=args.seqi))
        with open(args.seqi, "rt") as source:
            read_seqi_from_input_sources(source=source, delimiter=args.seqi_delimiter, destination=sequences_by_ids)
        logger.debug("Obtained {ms_cnt} meta-sequences records".format(ms_cnt=len(sequences_by_ids)))

    participating_sequences_by_ids = {}
    # for every fragment name that is not present as meta-sequence info a dumb one is created
    for fragment_aps in fragments:
        for f1, f1_or, f2, f2_or, gap_size in fragment_aps:
            if f1 not in sequences_by_ids:
                logger.debug("Fragment {f1} didn't have a meta-sequence information for it. Creating a dummy one.".format(f1=f1))
                sequences_by_ids[f1].append(Sequence(name=f1))
            if f2 not in sequences_by_ids:
                sequences_by_ids[f2].append(Sequence(name=f2))
                logger.debug("Fragment {f2} didn't have a meta-sequence information for it. Creating a dummy one.".format(f2=f2))
            participating_sequences_by_ids[f1] = sequences_by_ids[f1]
            participating_sequences_by_ids[f2] = sequences_by_ids[f2]

    all_seq_groups = set()
    for seq_id in list(participating_sequences_by_ids.keys()):
        for seq in participating_sequences_by_ids[seq_id]:
            all_seq_groups.add(seq.seq_group_id)
    if args.genomes_order == "":
        args.genomes_order = sorted(all_seq_groups, reverse=True)
    else:
        args.genomes_order = args.genomes_order.split(",")
    for genome in args.genomes_order:
        if genome not in all_seq_groups:
            logger.critical("Genome with name {genome} specified in the --genomes-order flag value was not present in the inferred set of sequences {seq_groups}"
                            "".format(genome=genome, seq_groups=",".join(sorted(all_seq_groups))))
            exit(1)
    # sorting participating sequences into "batches" denoted by seq_group_id field, with a default "None" value to it
    #   initial order of sequences within each batch is preserved w.r.t. the way it is listed in the *.seqi file
    for seq_id in list(participating_sequences_by_ids.keys()):
        in_genomes_order = [seq for seq in participating_sequences_by_ids[seq_id] if seq.seq_group_id in args.genomes_order]
        attributed_sequences = [seq for seq in participating_sequences_by_ids[seq_id] if seq.seq_group_id != "Default" and seq.seq_group_id not in args.genomes_order]
        non_attributed_sequences = [seq for seq in participating_sequences_by_ids[seq_id] if seq.seq_group_id == "Default"]
        participating_sequences_by_ids[seq_id] = []
        for genome in args.genomes_order:
            participating_sequences_by_ids[seq_id].extend([seq for seq in in_genomes_order if seq.seq_group_id == genome])
        attributed_sequences = sorted(attributed_sequences, reverse=True, key=lambda entry: entry.seq_group_id)
        participating_sequences_by_ids[seq_id].extend(attributed_sequences)
        participating_sequences_by_ids[seq_id].extend(non_attributed_sequences)

    meta_seqs_by_parent_ids = defaultdict(list)
    for seq_batch in participating_sequences_by_ids.values():
        for seq in seq_batch:
            meta_seqs_by_parent_ids[seq.parent_seq_id].append(seq)

    # for each parent sequence id all its "child" members are sorted with respect to their location on the parent sequence
    for parent_seq_id in list(meta_seqs_by_parent_ids.keys()):
        meta_seqs_by_parent_ids[parent_seq_id] = sorted(meta_seqs_by_parent_ids[parent_seq_id], key=lambda seq: (seq.start, seq.end))

    logger.info("Total number of fasta sequences is {seq_cnt}".format(seq_cnt=len(frag_fasta_by_id)))

    for seq_batch in participating_sequences_by_ids.values():
        for seq in seq_batch:
            if seq.parent_seq_id not in frag_fasta_by_id:
                logger.critical("Fragment {sequence} (parent for {block_name}) is not present in supplied fasta file. Exiting.".format(sequence=seq.parent_seq_id, block_name=seq.name))
                exit(1)

    used_fragments = set()
    filled_gaps_by_origin_cnt = Counter()
    extended_extremities_by_origin_cnt = Counter()
    total_gaps_cnt = 0
    total_extremities_cnt = 0
    for s_cnt, fragment_aps in enumerate(fragments):
        current = Seq("")
        total_extremities_cnt += 2
        for f_cnt, (f1, f1_or, f2, f2_or, gap_size) in enumerate(fragment_aps):
            main_meta_seq1 = participating_sequences_by_ids[f1][0]
            main_meta_seq2 = participating_sequences_by_ids[f2][0]

            seq1 = get_fasta_for_meta_seq(meta_seq=main_meta_seq1, orientation=f1_or, fasta_by_ids=frag_fasta_by_id)

            if f_cnt == 0 and args.extend_ends:
                logger.debug("Trying to extend the end for the new scaffold, based on fragment {f1}".format(f1=main_meta_seq1.name))
                flanking_meta_seqs = get_meta_flanking_by_meta_seq_name(meta_seq_name=main_meta_seq1.name, orientation=f1_or,
                                                                        meta_sequences_by_name=participating_sequences_by_ids, meta_sequences_by_parent_ids=meta_seqs_by_parent_ids,
                                                                        fasta_by_ids=frag_fasta_by_id, extension_type="before")
                if len(flanking_meta_seqs) > 0:
                    extension_meta_seq = flanking_meta_seqs[0]
                    extension = get_fasta_for_meta_seq(meta_seq=extension_meta_seq, orientation="+", fasta_by_ids=frag_fasta_by_id)
                    current += extension
                    extended_extremities_by_origin_cnt[extension_meta_seq.seq_group_id] += 1
                    logger.debug("Extended the end of the new scaffold based on the fragment {f1}, which appears to be the first in the \"chain\"".format(f1=f1))
                    logger.debug("\textension of length {ex_length} was obtained from the parent sequence {p_seq} in sequence group {seq_group_id}".format(ex_length=extension_meta_seq.length, p_seq=extension_meta_seq.parent_seq_id, seq_group_id=extension_meta_seq.seq_group_id))
                else:
                    logger.debug("Were unable to extend new scaffold based on the fragment {f1}, which appears to be the first in the \"chain\"".format(f1=f1))

            current += seq1
            total_gaps_cnt += 1
            filled_gap = False
            if args.fill_gaps:
                ap_gap_size = gap_size if isinstance(gap_size, numbers.Number) else "?"
                if ap_gap_size == "?" and not args.fill_gaps_unknown:
                    gap_filling = get_dummy_gap_filling(gap_size=ap_gap_size, sep=args.c_sep, default_gap_size=args.c_sep_length)
                    current += gap_filling
                    continue
                logger.debug("Trying to fill the gap for the new scaffold between fragments {f1} and {f2}".format(f1=f1, f2=f2))
                seq1_by_batches = defaultdict(list)
                seq1_batches_names = []
                for seq in participating_sequences_by_ids[f1]:
                    seq1_by_batches[seq.seq_group_id].append(seq)
                    if seq.seq_group_id not in seq1_batches_names:
                        seq1_batches_names.append(seq.seq_group_id)
                seq2_by_batches = defaultdict(list)
                seq2_batches_names = []
                for seq in participating_sequences_by_ids[f2]:
                    seq2_by_batches[seq.seq_group_id].append(seq)
                    if seq.seq_group_id not in seq2_batches_names:
                        seq2_batches_names.append(seq.seq_group_id)
                common_batches = [seq_group_id for seq_group_id in seq1_batches_names if seq_group_id in seq2_batches_names]
                logger.debug("{batches_cnt} sequence groups were identified where both {f1} and {f2} have sequence information".format(batches_cnt=len(common_batches), f1=f1, f2=f2))

                best_inter_gap_filling = None
                for seq_group_id in common_batches:
                    logger.debug("Trying to fill the gap using sequence group {seq_group_id}".format(seq_group_id=seq_group_id))
                    meta_seq1s = seq1_by_batches[seq_group_id]
                    meta_seq2s = seq2_by_batches[seq_group_id]
                    possible_fillings = []
                    for meta_seq1, meta_seq2 in itertools.product(meta_seq1s, meta_seq2s):
                        # case when both sequences reported in the assembly point have flanking sequences on their parent fragments
                        meta_seq1_flanking = get_meta_flanking_by_meta_seq(extension_type="after", fasta_by_ids=frag_fasta_by_id, meta_sequence=meta_seq1, meta_sequences_by_parent_ids=meta_seqs_by_parent_ids, orientation=f1_or)
                        meta_seq2_flanking = get_meta_flanking_by_meta_seq(extension_type="before", fasta_by_ids=frag_fasta_by_id, meta_sequence=meta_seq2, meta_sequences_by_parent_ids=meta_seqs_by_parent_ids, orientation=f2_or)
                        if meta_seq1_flanking is None or meta_seq2_flanking is None:
                            pass
                        else:
                            start_seq = get_fasta_for_meta_seq(meta_seq=meta_seq1_flanking, orientation="+", fasta_by_ids=frag_fasta_by_id)
                            end_seq = get_fasta_for_meta_seq(meta_seq=meta_seq2_flanking, orientation="+", fasta_by_ids=frag_fasta_by_id)
                            gap_filling = FlankingGapFilling(seq1=start_seq, seq2=end_seq, seq=None)
                            if args.gap_fill_inter:
                                possible_fillings.append(gap_filling)

                        # case when two sequences reported in the assembly point are on the same parent assembly and are co-oriented w.r.t. relative orientation reported in the assembly point
                        if meta_seq1.parent_seq_id == meta_seq2.parent_seq_id:
                            formation = collinearity(meta_seq1=meta_seq1, meta_seq2=meta_seq2, orientation1=f1_or, orientation2=f2_or)
                            if formation == 0:
                                continue
                            if formation == 1:
                                meta_seq_filling = Sequence(name="filling", parent_seq_id=meta_seq1.parent_seq_id, start=meta_seq1.end, end=meta_seq2.start, strand="+", seq_group_id=seq_group_id)
                            else:
                                meta_seq_filling = Sequence(name="filling", parent_seq_id=meta_seq1.parent_seq_id, start=meta_seq2.end, end=meta_seq1.start, strand="-", seq_group_id=seq_group_id)
                            filling_seq = get_fasta_for_meta_seq(meta_seq=meta_seq_filling, orientation="+", fasta_by_ids=frag_fasta_by_id)
                            gap_filling = IntraGapFilling(seq=filling_seq)
                            possible_fillings.append(gap_filling)
                    if len(possible_fillings) == 0:
                        logger.debug("Failed to fill the gap using sequence group {seq_group_id}".format(seq_group_id=seq_group_id))
                    else:
                        suitable_gap_fillings = []
                        per_based_gap_error = (ap_gap_size / 100 * args.gap_diff_threshold_per) if ap_gap_size != "?" else 0
                        bp_based_gap_error = args.gap_diff_threshold_bp if ap_gap_size != "?" else 0
                        gap_error = max(per_based_gap_error, bp_based_gap_error) if args.gap_diff_max_over_min else min(per_based_gap_error, bp_based_gap_error)
                        if ap_gap_size != "?":
                            gap_error = min(ap_gap_size, gap_error)
                            logger.debug("Inferred gap size error is {gse} for the gap of size {gs}".format(gse=gap_error, gs=ap_gap_size))
                        for gap_filling in possible_fillings:
                            if isinstance(gap_filling, IntraGapFilling) and gap_filling.suitable_to_fill_the_gap(gap_size=ap_gap_size, gap_size_error=gap_error):
                                gap_filling.compute_score(gap_size=ap_gap_size, gap_size_error=gap_error)
                                suitable_gap_fillings.append(gap_filling)
                            elif isinstance(gap_filling, FlankingGapFilling) and gap_filling.suitable_to_fill_the_gap(gap_size=ap_gap_size, gap_size_error=gap_error):
                                gap_filling.prepare_seq(gap_size=ap_gap_size, gap_size_error=gap_error)
                                gap_filling.compute_score(gap_size=ap_gap_size, gap_size_error=gap_error)
                                suitable_gap_fillings.append(gap_filling)
                        if len(suitable_gap_fillings) == 0:
                            logger.debug("Failed to fill the gap using sequence group {seq_group_id}".format(seq_group_id=seq_group_id))
                        else:
                            logger.debug("Filled the gap between fragments {f1} and {f2} using sequence group {seq_group_id}".format(f1=f1, f2=f2, seq_group_id=seq_group_id))
                            if args.gap_fill_intra_first:
                                intra_gap_fillings = [filling for filling in suitable_gap_fillings if isinstance(filling, IntraGapFilling)]
                                if len(intra_gap_fillings) > 0:
                                    gap_filling = sorted(suitable_gap_fillings, key=lambda entry: entry.score)[-1]
                                    current += gap_filling.seq
                                    filled_gap = True
                                    filled_gaps_by_origin_cnt[seq_group_id] += 1
                                else:
                                    if best_inter_gap_filling is None:
                                        best_inter_gap_filling = sorted(suitable_gap_fillings, key=lambda entry: entry.score)[-1]
                            else:
                                gap_filling = sorted(suitable_gap_fillings, key=lambda entry: entry.score)[-1]
                                current += gap_filling.seq
                                filled_gap = True
                                filled_gaps_by_origin_cnt[seq_group_id] += 1
                    if filled_gap:
                        break
                # is only triggered if the gap was NOT successfully filled with a non "N" sequence
                else:
                    if best_inter_gap_filling is not None:
                        gap_filling = best_inter_gap_filling.seq
                    else:
                        gap_filling = get_dummy_gap_filling(gap_size=ap_gap_size, sep=args.c_sep, default_gap_size=args.c_sep_length)
                    current += gap_filling

            if f_cnt == len(fragment_aps) - 1:
                seq2 = get_fasta_for_meta_seq(meta_seq=main_meta_seq2, orientation=f2_or, fasta_by_ids=frag_fasta_by_id)

                current += seq2

                if args.extend_ends:
                    logger.debug("Trying to extend the end for the new scaffold, based on fragment {f2}".format(f2=f2))
                    flanking_meta_seqs = get_meta_flanking_by_meta_seq_name(meta_seq_name=main_meta_seq2.name, orientation=f2_or,
                                                                            meta_sequences_by_name=participating_sequences_by_ids, meta_sequences_by_parent_ids=meta_seqs_by_parent_ids,
                                                                            fasta_by_ids=frag_fasta_by_id, extension_type="after")
                    if len(flanking_meta_seqs) > 0:
                        extension_meta_seq = flanking_meta_seqs[0]
                        extension = get_fasta_for_meta_seq(meta_seq=extension_meta_seq, orientation="+", fasta_by_ids=frag_fasta_by_id)
                        current += extension
                        extended_extremities_by_origin_cnt[extension_meta_seq.seq_group_id] += 1
                        logger.debug("Extended the end of the new scaffold based on the fragment {f2}, which appears to be the last in the \"chain\"".format(f2=f2))
                        logger.debug("\textension of length {ex_length} was obtained from the parent sequence {p_seq} in sequence group {seq_group_id}".format(ex_length=extension_meta_seq.length, p_seq=extension_meta_seq.parent_seq_id, seq_group_id=extension_meta_seq.seq_group_id))
                    else:
                        logger.debug("Were unable to extend new scaffold based on the fragment {f2}, which appears to be the first in the \"chain\"".format(f2=f2))

        name = args.scaffold_name_template.format(cnt=s_cnt)
        seq_record = SeqRecord(seq=current, id=name, description="")
        SeqIO.write(sequences=seq_record, handle=args.output, format="fasta")
    logger.info("Filled {gap_fill_cnt} / {total_gaps_cnt} gaps.".format(gap_fill_cnt=sum(filled_gaps_by_origin_cnt.values()), total_gaps_cnt=total_gaps_cnt))
    for key, value in filled_gaps_by_origin_cnt.items():
        logger.debug("\tFilled {gap_filled_cnt} gaps with source \"{seq_group_id}\"".format(gap_filled_cnt=value, seq_group_id=key))
    logger.info("Extended {end_ext_cnt} / {total_extremities_cnt} extremities.".format(end_ext_cnt=sum(extended_extremities_by_origin_cnt.values()), total_extremities_cnt=total_extremities_cnt))
    for key, value in extended_extremities_by_origin_cnt.items():
        logger.debug("\tExtended {extended_extremities_cnt} extremities of new scaffolds with source \"{seq_group_id}\"".format(extended_extremities_cnt=value, seq_group_id=key))
    # if args.allow_singletons:
    #     logger.info("Adding singleton fragments, that did not participate in any assembly points to the resulting assembly")
    #     for f_id, fragment in frag_fasta_by_id.items():
    #         if f_id not in used_fragments:
    #             SeqIO.write(sequences=fragment, handle=args.output, format="fasta")
    logger.info("New scaffolds were written to {file_name}".format(file_name=args.output))
    logger.info("All done!")
    logger.info("Elapsed time: {el_time}".format(el_time=str(datetime.datetime.now() - start_time)))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import os
import sys
from collections import defaultdict

import configargparse
import logging
from bg.grimm import GRIMMReader

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
from camsa.core.data_structures import AssemblyPoint
import camsa.core.io as camsa_io


def get_assembly_points(genomes):
    result = []
    for genome, chromosomes in genomes.items():
        for chr_type, blocks in chromosomes:
            left_blocks, right_blocks = blocks[:-1], blocks[1:]
            for l_block, r_block in zip(left_blocks, right_blocks):
                l_block_sign, l_block_name = l_block
                r_block_sign, r_block_name = r_block
                ap = AssemblyPoint(seq1=l_block_name, seq2=r_block_name, seq1_or=l_block_sign, seq2_or=r_block_sign, sources=[genome])
                result.append(ap)
            if chr_type == "@":
                first_block_sign, first_block_name = blocks[0]
                last_block_sign, last_block_name = blocks[-1]
                ap = AssemblyPoint(seq1=last_block_name, seq1_or=last_block_sign, seq2=first_block_name, seq2_or=first_block_sign, sources=[genome])
                result.append(ap)
    return result


def get_blocks_cnt(genomes):
    block_counter = defaultdict(lambda: defaultdict(int))
    for genome, chromosomes in genomes.items():
        for _, blocks in chromosomes:
            for block_sign, block_name in blocks:
                block_counter[block_name][genome] += 1
    return block_counter


def remove_bad_blocks(genomes, bad_blocks):
    for genome_name in list(genomes.keys()):
        new_chromosomes = []
        for chr_type, old_blocks in genomes[genome_name]:
            new_blocks = []
            for block_sign, block_name in old_blocks:
                if block_name in bad_blocks:
                    continue
                new_blocks.append((block_sign, block_name))
            if len(new_blocks) != 0:
                new_chromosomes.append((chr_type, new_blocks))
        genomes[genome_name] = new_chromosomes


def remove_duplications(genomes):
    blocks_cnt = get_blocks_cnt(genomes=genomes)
    bad_blocks = set()
    for block_name, block_cnt in blocks_cnt.items():
        cnt_per_genomes = block_cnt.values()
        if any(map(lambda x: x > 1, cnt_per_genomes)):
            bad_blocks.add(block_name)
    remove_bad_blocks(genomes=genomes, bad_blocks=bad_blocks)


def remove_indels(genomes):
    blocks_cnt = get_blocks_cnt(genomes=genomes)
    bad_blocks = set()
    all_genomes = set(genomes.keys())
    for block_name, block_cnt in blocks_cnt.items():
        blocks_genomes = set(block_cnt.keys())
        if len(all_genomes - blocks_genomes) != 0:
            bad_blocks.add(block_name)
    remove_bad_blocks(genomes=genomes, bad_blocks=bad_blocks)

if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting GRIMM formatted scaffolding results for further CAMSA processing.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "utils", "grimm", "grimm2camsa_points.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])
    parser.add_argument("grimm", nargs="+", help="A list of paths to files, that in GRIMM format contain information about scaffold assemblies.")
    parser.add_argument("--o-format", type=str, help="")
    parser.add_argument("-o", "--output", type=configargparse.FileType("wt"), default=sys.stdout, help="Path to the file, where converted assembly points will be stored.\nDEFAULT: stdout")
    parser.add_argument("--o-delimiter", default="\t", type=str, help="")
    parser.add_argument("--no-trim-names", dest="trim_names", default=True, action="store_false",
                        help="A flag to indicate, that genome names from grimm file need not to be trimmed by the first \".\"\nDEFAULT: true")
    parser.add_argument("--trimmer-char", default=".", type=str, help="A character, which first occurrence in the genome name indicates a position for trimming.\nDEFAULT: .")
    parser.add_argument("--good-genomes", type=str, default="", help="A coma separated list of genome names, to be processed and conversed.\nDEFAULT: \"\" (i.e., all genomes are good)")
    parser.add_argument("--bad-genomes", type=str, default="", help="A coma separated list of genome names, to be excluded from processing and conversion.\nDEFAULT: \"\" (i.e., no genomes are bad)")
    parser.add_argument("--filter-duplications", dest="filter_duplications", default=False, action="store_true", help="")
    parser.add_argument("--filter-indels", dest="filter_indels", default=False, action="store_true", help="")
    parser.add_argument("--c-logging-level", type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")

    args = parser.parse_args()

    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.grimm2camsa_points")
    ch = logging.StreamHandler()
    ch.setLevel(args.c_logging_level)
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    genomes = defaultdict(list)
    for file_name in args.grimm:
        logger.info("Processing file \"{file_name}\"".format(file_name=file_name))
        with open(file_name, "rt") as source:
            current_genome = None
            for line in source:
                line = line.strip()
                if len(line) == 0 or GRIMMReader.is_comment_string(data_string=line):
                    continue
                if GRIMMReader.is_genome_declaration_string(data_string=line):
                    current_genome = GRIMMReader.parse_genome_declaration_string(data_string=line).name
                    if args.trim_names:
                        current_genome = current_genome.split(args.trimmer_char, 1)[0]
                elif current_genome is not None:
                    current_chromosome = []
                    chr_type, blocks = GRIMMReader.parse_data_string(data_string=line)
                    genomes[current_genome].append((chr_type, blocks))
    if args.good_genomes != "":
        good_genomes = args.good_genomes.split(",")
        if args.trim_names:
            good_genomes = [genome_name.split(args.trimmer_char, 1)[0] for genome_name in good_genomes]
        for genome_name in list(genomes.keys()):
            if genome_name not in good_genomes:
                del genomes[genome_name]
    if args.bad_genomes != "":
        bad_genomes = args.bad_genomes.split(",")
        if args.trim_names:
            bad_genomes = [genome_name.split(args.trimmer_char, 1)[0] for genome_name in bad_genomes]
        for genome_name in list(genomes.keys()):
            if genome_name in bad_genomes:
                del genomes[genome_name]
    if args.filter_duplications:
        remove_duplications(genomes=genomes)
    result = get_assembly_points(genomes=genomes)
    logger.info("Writing output to file \"{file_name}\"".format(file_name=args.output))
    camsa_io.write_assembly_points(assembly_points=result, destination=args.output, output_setup=args.o_format, delimiter=args.o_delimiter)
    logger.info("Elapsed time: {el_time}".format(el_time=str(datetime.datetime.now() - start_time)))






